"""Dataclasses for experiment and it's components"""

from dataclasses import dataclass, field
from datetime import datetime
from typing import List
from uuid import UUID


@dataclass
class ExperimentFile:
    """Dataclass for experiment file"""

    name: str
    path: str
    modified_at: datetime

    @classmethod
    def from_dict(cls, data):
        """Convert experiment file object to dictionary"""
        return cls(
            name=data["name"],
            path=data["path"],
            modified_at=datetime.fromisoformat(data["modifiedAt"]),
        )


@dataclass
class ExperimentData:
    """Dataclass for experiment"""

    uuid: UUID
    title: str
    description: str
    eid: str
    created_at: datetime
    updated_at: datetime
    tags: List[str] = field(default_factory=list)
    files: List[ExperimentFile] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data):
        """Convert experiment object to dictionary"""
        return cls(
            uuid=UUID(data["uuid"]),
            title=data["title"],
            description=data["description"],
            tags=data["tags"],
            eid=data["eid"],
            files=[ExperimentFile.from_dict(file_data) for file_data in data["files"]],
            created_at=datetime.fromisoformat(data["createdAt"]),
            updated_at=datetime.fromisoformat(data["updatedAt"]),
        )


@dataclass
class ExperimentsInfo:
    """Dataclass for experiment info"""

    experiments: List[ExperimentData]
    total_count: int

    @classmethod
    def from_dict(cls, data):
        """Convert experiment info object to dictionary"""
        return cls(
            experiments=[
                ExperimentData.from_dict(exp_data) for exp_data in data["experimentsData"]
            ],
            total_count=data["totalExperimentsCount"],
        )


@dataclass
class TagsData:
    """Dataclass for definition of tag"""

    tags: List[str]
    total_count: int

    @classmethod
    def from_dict(cls, data):
        """Convert tag data class to a dictionary"""
        return cls(tags=data["tagsData"], total_count=data["totalTagsCount"])



"""Aqueduct exceptions."""


class PyAqueductError(Exception):
    """Aqueduct exception base class."""


class RemoteOperationError(PyAqueductError):
    """Remote operation error."""


class ForbiddenError(PyAqueductError):
    """Operation forbidden error."""


class UnAuthorizedError(PyAqueductError):
    """Unauthorized operation error."""


class FileToUploadNotFoundError(PyAqueductError):
    """File not found error."""


class FileUploadError(PyAqueductError):
    """Interrupted upload error."""


class FileDownloadError(PyAqueductError):
    """Interrupted download error."""



"""Experiment module."""

from __future__ import annotations

from datetime import datetime
from typing import List, Tuple
from uuid import UUID

from pydantic import BaseModel, validate_call

from pyaqueduct.client import AqueductClient


class Experiment(BaseModel):
    """Experiment model."""

    _client: AqueductClient
    "Client object reference."
    uuid: UUID
    "UUID of the experiment. This is an internal experiment identifier in the database"
    eid: str
    "EID of the experiment. User-readable identifier, it is unique within one Aqueduct installation"
    created_at: datetime
    "Creation datetime of the experiment."

    def __init__(self, uuid: UUID, eid: str, created_at: datetime, client: AqueductClient, **data):
        super().__init__(uuid=uuid, eid=eid, created_at=created_at, **data)
        self._client = client

    @property
    def title(self) -> str:
        """Get title of experiment."""
        return self._client.get_experiment(experiment_uuid=self.uuid).title

    @title.setter
    @validate_call
    def title(self, value: str) -> None:
        """Set title of experiment.

        Args:
            value: New title.

        """
        self._client.update_experiment(experiment_uuid=self.uuid, title=value)

    @property
    def description(self) -> str:
        """Get description of experiment."""
        return self._client.get_experiment(self.uuid).description

    @description.setter
    @validate_call
    def description(self, value: str) -> None:
        """Set description of experiment.

        Args:
            value: New description.

        """
        self._client.update_experiment(experiment_uuid=self.uuid, description=value)

    @property
    def tags(self) -> List[str]:
        """Gets tags of experiment."""
        return self._client.get_experiment(self.uuid).tags

    @validate_call
    def add_tags(self, tags: List[str]) -> None:
        """Add new tags to experiment.

        Args:
            tags: List of tags to be added to the experiment.

        """
        self._client.add_tags_to_experiment(experiment_uuid=self.uuid, tags=tags)

    @validate_call
    def remove_tag(self, tag: str) -> None:
        """Remove tag from experiment."""
        self._client.remove_tag_from_experiment(experiment_uuid=self.uuid, tag=tag)

    @property
    def files(self) -> List[Tuple[str, datetime]]:
        """Get file names of expriment."""
        return [
            (item.name, item.modified_at) for item in self._client.get_experiment(self.uuid).files
        ]

    @validate_call
    def download_file(self, file_name: str, destination_dir: str) -> None:
        """Download the specified file of experiment."""
        self._client.download_file(self.uuid, file_name=file_name, destination_dir=destination_dir)

    @validate_call
    def upload_file(self, file: str) -> None:
        """Upload the specified file to experiment."""
        self._client.upload_file(self.uuid, file=file)

    @property
    def updated_at(self) -> datetime:
        """Get last updated datetime of the experiment."""
        return self._client.get_experiment(self.uuid).updated_at



""" The module contains classes representing interface with extensions. Extension
list may be retrieved from the server using api method `API.get_extensions()`.
Each extension may have one or more actions. Each action has a list of expected
parameters."""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from pyaqueduct.client import AqueductClient
from pyaqueduct.client.extension_types import (ExtensionExecutionResultData,
                                            ExtensionActionData,
                                            ExtensionParameterData)


class ExtensionAction(BaseModel):
    """Extension action representation. Contains an execution method
    which trigger extension action execution on the side of Aqueduct server."""

    parameters: List[ExtensionParameterData]
    """List of parameters which extension action expects to accept."""

    data: ExtensionActionData
    extension: Extension = None
    """Extension to which this action belongs."""

    _client: AqueductClient = None

    def __init__(
        self,
        extension: Extension,
        action_data: ExtensionAction,
        client: AqueductClient,
    ):
        super().__init__(extension=extension, data=action_data, parameters=action_data.parameters)
        self._client = client

    @property
    def name(self) -> str:
        """Extension action name. Unique inside an extension."""
        return self.data.name

    @property
    def description(self) -> str:
        """Detailed description of the extension action."""
        return self.data.description

    @property
    def experiment_variable_name(self) -> str:
        """Name of the variable which is used to define a default experiment.
        This experiment will be used to save logs and validate variables
        of `file` type."""
        return self.data.experimentVariableName

    def execute(self, parameters: Dict[str, Any]) -> ExtensionExecutionResultData:
        """Execute an extension action on a server.

        Args:
            parameters: dictionary of parameters to pass to an extension.

        Returns:
            result of extension execution on server. `returnCode==0` corresponds to success.
        """
        return self._client.execute_extension_action(
            extension=self.extension.name,
            action=self.data.name,
            params=parameters,
        )


class Extension(BaseModel):
    """Class represents an extension as a collection of actions."""

    name: str
    """Extension name. Unique name within a server"""

    description: Optional[str]
    """Description of extension scope and overview of its actions."""

    authors: str
    """Authors of the extension."""

    actions: List[ExtensionAction]

    def __init__(self, name: str, description: Optional[str],
                 authors: str, actions: List[ExtensionActionData],
                 client: AqueductClient):
        super().__init__(name=name, description=description,
                         authors=authors, actions=[])
        for action in actions:
            self.actions.append(ExtensionAction(self, action, client))



"""PyAqueduct Settings."""

from typing import Optional

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    """Settings class to be read from the environment variables."""

    api_token: Optional[str] = None



"""Aqueduct client module to communicate with the server instance."""

from pyaqueduct.client.client import AqueductClient
from pyaqueduct.client.experiment_types import ExperimentData, ExperimentFile, ExperimentsInfo
from pyaqueduct.client.extension_types import (
    ExtensionData, ExtensionActionData, ExtensionExecutionResultData, ExtensionParameterData)

__all__ = ["AqueductClient", "ExperimentData", "ExperimentFile", "ExperimentsInfo",
           "ExtensionData", "ExtensionActionData", "ExtensionExecutionResultData", 
           "ExtensionParameterData"]



"""Aqueduct application programming interface (API) module."""

from __future__ import annotations

from datetime import datetime
from uuid import UUID
from typing import List, Optional

from pydantic import (
    BaseModel,
    HttpUrl,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    PrivateAttr,
    validate_call,
)

from pyaqueduct.client import AqueductClient
from pyaqueduct.experiment import Experiment
from pyaqueduct.extensions import Extension
from pyaqueduct.settings import Settings


class API(BaseModel):
    """Aqueduct API interface to interact with experiments.

    Args:
        url: URL of the Aqueduct server including the prefix.
        timeout: Timeout of operations in seconds.

    """

    url: HttpUrl
    timeout: PositiveFloat

    _client: AqueductClient = PrivateAttr()

    def __init__(self, url: str, timeout: float = 0.5):
        super().__init__(url=url, timeout=timeout)

        if not url.endswith("/"):
            url = url + "/"
        api_url = f"{url}api"

        self._client = AqueductClient(url=api_url, timeout=timeout, api_token=Settings().api_token)

    @validate_call
    def create_experiment(
        self,
        title: str,
        description: str,
    ) -> Experiment:
        """Create an experiment with specific title and description.

        Args:
            title: Title of the experiment.
            description: Description of the experiment.

        Returns:
            Experiment object to interact with its data.

        """

        experiment_data = self._client.create_experiment(title=title, description=description)
        return Experiment(
            client=self._client,
            uuid=experiment_data.uuid,
            eid=experiment_data.eid,
            created_at=experiment_data.created_at,
        )

    @validate_call
    def get_experiment_by_eid(self, eid: str) -> Experiment:
        """Get the experiment by the specified identifier to operate on.

        Args:
            eid: EID of the specified experiment.

        Returns:
            Experiment object to interact with the experiment data.

        """
        experiment_data = self._client.get_experiment_by_eid(eid=eid)
        return Experiment(
            client=self._client,
            uuid=experiment_data.uuid,
            eid=experiment_data.eid,
            created_at=experiment_data.created_at,
        )

    @validate_call
    def get_experiment_by_uuid(self, uuid: UUID) -> Experiment:
        """Get the experiment by the specified identifier to operate on.

        Args:
            uuid: UUID of the specified experiment.

        Returns:
            Experiment object to interact with the experiment data.

        """
        experiment_data = self._client.get_experiment(experiment_uuid=uuid)
        return Experiment(
            client=self._client,
            uuid=experiment_data.uuid,
            eid=experiment_data.eid,
            created_at=experiment_data.created_at,
        )

    @validate_call
    def remove_experiment_by_eid(self, eid: str) -> None:
        """Remove experiment from the database. Experiment's files will be also removed.

        Args:
            eid: EID of the specified experiment.

        """
        experiment_data = self._client.get_experiment_by_eid(eid=eid)
        self._client.remove_experiment(experiment_uuid=experiment_data.uuid)

    @validate_call
    def find_experiments(
        self,
        search: Optional[str] = None,
        limit: PositiveInt = 10,
        offset: NonNegativeInt = 0,
        tags: Optional[List[str]] = None,
        start_datetime: Optional[datetime] = None,
        end_datetime: Optional[datetime] = None,
    ) -> List[Experiment]:
        """Find the experiments that have the search criteria provided in arguments.

        Args:
            search: The string to search for in the title field of experiments.
            limit: The maximum number of experiments to fetch in a single request.
            offset: The number of experiments to skip from the beginning of the search results.
            tags: List of tags to filter the experiments by.
            start_datetime: Start datetime to filter the experiments after this date and time.
            end_datetime: End datetime to filter the experiments before this date and time.

        Returns:
            List of experiment objects to operate on their data.

        """
        experiments = self._client.get_experiments(
            title=search,
            limit=limit,
            offset=offset,
            tags=tags,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        ).experiments
        return [
            Experiment(
                client=self._client,
                uuid=experiment.uuid,
                eid=experiment.eid,
                created_at=experiment.created_at,
            )
            for experiment in experiments
        ]

    @validate_call
    def get_extensions(self) -> List[Extension]:
        """Gets the current fresh extension list from the server.
        Extension list may change without server restart.

        Returns:
            List of extension objects.
        """
        return [
            Extension(
                name=extension_data.name,
                description=extension_data.description,
                authors=extension_data.authors,
                actions=extension_data.actions,
                client=self._client,
            )
            for extension_data in self._client.get_extensions()
        ]



"""Aqueduct client class to enable experiment based operations."""

import logging
import os
from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import UUID

from gql import Client
from gql.client import SyncClientSession
from gql.transport import exceptions as gql_exceptions
from gql.transport.httpx import HTTPXTransport
from graphql import DocumentNode
from httpx import TransportError, codes, post, stream
from pydantic import BaseModel, HttpUrl, PrivateAttr
from tqdm import tqdm

from pyaqueduct.client.experiment_types import ExperimentData, ExperimentsInfo, TagsData
from pyaqueduct.client.extension_types import ExtensionData, ExtensionExecutionResultData
from pyaqueduct.exceptions import (
    FileDownloadError,
    FileUploadError,
    ForbiddenError,
    RemoteOperationError,
    UnAuthorizedError,
)
from pyaqueduct.schemas.mutations import (
    add_tags_to_experiment_mutation,
    create_experiment_mutation,
    remove_experiment_mutation,
    execute_extension_action_mutation,
    remove_tag_from_experiment_mutation,
    update_experiment_mutation,
)
from pyaqueduct.schemas.queries import (
    get_all_extensions_query,
    get_all_tags_query,
    get_experiment_query,
    get_experiments_query,
)


def process_response_common(code: codes) -> None:
    """Process common HTTP return codes."""
    if code is codes.OK:
        return

    if code is codes.FORBIDDEN:
        raise ForbiddenError("Operation is not allowed for the current user.") from None

    if code is codes.UNAUTHORIZED:
        raise UnAuthorizedError("API token couldn't be verified or is missing.") from None

    raise RemoteOperationError("Remove operation failed.")


class AqueductClient(BaseModel):
    """
    AqueductClient - A client class for managing experiments, tags and files.

    This class provides methods for creating experiments, updating experiments,
    adding and removing tags, as well as uploading and downloading files.

    """

    url: HttpUrl
    timeout: float
    _gql_client: Client = PrivateAttr()
    _session: SyncClientSession = PrivateAttr()
    _headers: Dict[str, str] = PrivateAttr()

    def __init__(self, url: str, timeout: float, api_token: Optional[str] = None):
        """
        Args:
            url: URL of the Aqueduct server endpoint.
            timeout: Response timeout in seconds.

        """
        super().__init__(url=url, timeout=timeout)
        self._headers = {"Authorization": f"Bearer {api_token}"} if api_token else {}

        self._gql_client = Client(
            transport=HTTPXTransport(
                url=f"{url}/graphql", timeout=self.timeout, headers=self._headers
            )
        )

    def fetch_response(self, operation: DocumentNode, variable_values: Dict) -> Dict[str, Any]:
        """
        Send query or mutation request to the server.

        Args:
            operation: Query or mutation schema.
            variable_values: Values for the params to be sent in the request.

        Returns:
            A JSON object
        """
        try:
            data = self._gql_client.execute(
                operation,
                variable_values=variable_values,
            )
        except gql_exceptions.TransportServerError as error:
            if error.code:
                process_response_common(codes(error.code))
            raise
        except gql_exceptions.TransportQueryError as error:
            raise RemoteOperationError(
                error.errors if error.errors else "Unknown error occurred in the remote operation."
            ) from error
        return data

    def create_experiment(
        self, title: str, description: str, tags: Optional[List[str]] = None
    ) -> ExperimentData:
        """
        Create an experiment with title, description and list of tags.

        Args:
            title: Title of experiment.
            description: Description of experiment.
            tags: List of tags to be assigned to experiment.

        Returns:
            Experiment object.
        """

        data = self.fetch_response(
            create_experiment_mutation,
            {"title": title, "description": description, "tags": tags or []},
        )
        experiment_obj = ExperimentData.from_dict(
            data["createExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Created experiment - %s - %s", experiment_obj.uuid, experiment_obj.title)
        return experiment_obj

    def update_experiment(
        self, experiment_uuid: UUID, title: Optional[str] = None, description: Optional[str] = None
    ) -> ExperimentData:
        """
        Update title or description or both for experiment.

        Args:
            experiment_uuid: UUID of experiment.
            title: New title of experiment.
            description: New description of experiment.

        Returns:
            Experiment object.

        """
        data = self.fetch_response(
            update_experiment_mutation,
            {
                "uuid": str(experiment_uuid),
                "title": title,
                "description": description,
            },
        )
        experiment_obj = ExperimentData.from_dict(
            data["updateExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Updated experiment - %s", experiment_obj.uuid)
        return experiment_obj

    def get_experiments(
        self,
        limit: int,
        offset: int,
        title: Optional[str] = None,
        tags: Optional[List[str]] = None,
        start_datetime: Optional[datetime] = None,
        end_datetime: Optional[datetime] = None,
    ) -> ExperimentsInfo:
        """
        Get a list of experiments

        Args:
            limit: Pagination field, number of experiments to be fetched.
            offset: Pagination field, number of experiments to skip.
            title: Perform search on experiments through their title and EID.
            tags: Get experiments that have these tags.
            start_date: Start datetime to filter experiments (timezone aware).
            end_date: End datetime to filter experiments to (timezone aware).

        Returns:
            List of experiments with filters applied.

        """
        data = self.fetch_response(
            get_experiments_query,
            {
                "limit": limit,
                "offset": offset,
                "title": title,
                "startDate": start_datetime.isoformat() if start_datetime else None,
                "endDate": end_datetime.isoformat() if end_datetime else None,
                "tags": tags,
            },
        )
        experiments_obj = ExperimentsInfo.from_dict(
            data["experiments"]  # pylint: disable=unsubscriptable-object
        )
        logging.info(
            "Fetched %s experiments, total %s experiments",
            len(experiments_obj.experiments),
            experiments_obj.total_count,
        )
        return experiments_obj

    def get_experiment(self, experiment_uuid: UUID) -> ExperimentData:
        """
        Get an Experiment by UUID.

        Args:
            experiment_uuid: UUID of experiment.

        Returns:
            Experiment object.

        """
        data = self.fetch_response(
            get_experiment_query,
            {"type": "UUID", "value": str(experiment_uuid)},
        )
        experiment_obj = ExperimentData.from_dict(
            data["experiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Fetched experiment - %s", experiment_obj.title)
        return experiment_obj

    def get_experiment_by_eid(self, eid: str) -> ExperimentData:
        """
        Get an experiment by its EID.

        Args:
            EID: Experiment's EID.

        Returns:
            Updated experiment.

        """
        data = self.fetch_response(
            get_experiment_query,
            {"type": "EID", "value": eid},
        )
        experiment_obj = ExperimentData.from_dict(
            data["experiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Fetched experiment - %s", experiment_obj.title)
        return experiment_obj

    def add_tags_to_experiment(self, experiment_uuid: UUID, tags: List[str]) -> ExperimentData:
        """
        Add tags to an experiment.

        Args:
            experiment_uuid: UUID of experiment.
            tags: List of tags to be added to experiment.

        Returns:
            Updated experiment.

        """
        data = self.fetch_response(
            add_tags_to_experiment_mutation,
            {"uuid": str(experiment_uuid), "tags": tags},
        )
        experiment_obj = ExperimentData.from_dict(
            data["addTagsToExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Added tags %s to experiment <%s>", tags, experiment_obj.title)
        return experiment_obj

    def remove_experiment(self, experiment_uuid: UUID) -> None:
        """
        Remove experiment from the database. It removes the experiments files as well.

        Args:
            experiment_uuid: UUID of experiment.

        """
        self.fetch_response(
            remove_experiment_mutation,
            {"uuid": str(experiment_uuid)},
        )

    def remove_tag_from_experiment(self, experiment_uuid: UUID, tag: str) -> ExperimentData:
        """
        Remove a tag from an experiment

        Args:
        - experiment_uuid (UUID): UUID of experiment frin which tag has to be removed
        - tag (str): Tag to be removed from experiment

        Returns:
        Experiment: Experiment having tag removed
        """
        data = self.fetch_response(
            remove_tag_from_experiment_mutation,
            {"uuid": str(experiment_uuid), "tag": tag},
        )

        experiment_obj = ExperimentData.from_dict(
            data["removeTagFromExperiment"]  # pylint: disable=unsubscriptable-object
        )
        logging.info("Removed tag %s from experiment <%s>", tag, experiment_obj.title)
        return experiment_obj

    def get_tags(self, limit: int, offset: int, dangling: bool = True) -> TagsData:
        """
        Get a list of existing tags

        Args:
        - limit (int): Number of tags to be fetched
        - offset (offset): Number of tags to skip
        - dangling (bool): If tags not linked to any experiment should be included or not

        Returns:
        List[str]: A list of all existing tags
        """
        data = self.fetch_response(
            get_all_tags_query,
            {"limit": limit, "offset": offset, "dangling": dangling},
        )
        tags_obj = TagsData.from_dict(data["tags"])  # pylint: disable=unsubscriptable-object
        logging.info("Fetched %s tags, total %s tags", len(tags_obj.tags), tags_obj.total_count)
        return tags_obj

    def upload_file(self, experiment_uuid: UUID, file: str) -> None:
        """
        Upload file to a specific experiment.

        Args:
            experiment_uuid : The ID of the experiment.
            file_path: The local path to the file to be uploaded.

        Returns:
            Operation results object.

        """

        headers = {"file_name": os.path.basename(file), **self._headers}

        upload_url = f"{self.url}/files/{str(experiment_uuid)}"
        with open(file, "rb") as files:
            try:
                response = post(
                    upload_url, headers=headers, timeout=self.timeout, files={"file": files}
                )
            except TransportError as error:
                raise FileUploadError(f"Couldn't upload {file} due to transport error.") from error

        process_response_common(codes(response.status_code))

        logging.info("Successfully uploaded file {files}")

    def download_file(self, experiment_uuid: UUID, file_name: str, destination_dir: str) -> None:
        """
        Download file from a specific experiment.

        Args:
            experiment_uuid: The ID of the experiment.
            file_id: The ID of the file to be downloaded.
            destination: The local path where the downloaded file will be saved.

        Returns:
            Operation results object.

        """
        download_url: str = f"{self.url}/files/{experiment_uuid}/{file_name}"
        destination = f"{destination_dir}/{file_name}"

        try:
            with open(destination, "wb") as download_file:
                with stream(
                    "GET", download_url, timeout=self.timeout, headers=self._headers
                ) as response:
                    total = int(response.headers["Content-Length"])
                    with tqdm(
                        total=total, unit_scale=True, unit_divisor=1024, unit="B"
                    ) as progress:
                        num_bytes_downloaded = response.num_bytes_downloaded
                        for chunk in response.iter_bytes():
                            download_file.write(chunk)
                            progress.update(response.num_bytes_downloaded - num_bytes_downloaded)
                            num_bytes_downloaded = response.num_bytes_downloaded

        except Exception as error:
            raise FileDownloadError(
                f"Couldn't download {file_name} due to transport error."
            ) from error

    def get_extensions(self) -> List[ExtensionData]:
        """ Get the list of extensions from the server.

        Returns:
            List of extension objects.
        """
        try:
            extensions = self._gql_client.execute(
                get_all_extensions_query,
            )
        except gql_exceptions.TransportServerError as error:
            if error.code:
                process_response_common(codes(error.code))
            raise
        except gql_exceptions.TransportQueryError as error:
            raise RemoteOperationError(
                error.errors if error.errors else "Unknown error occurred in the remote operation."
            ) from error

        extensions = list(map(ExtensionData.from_dict, extensions["extensions"]))  # pylint: disable=unsubscriptable-object
        logging.info("Fetched %s extensions", len(extensions))
        return extensions

    def execute_extension_action(
            self, extension: str, action: str, params: Dict[str, Any]
        ) -> ExtensionExecutionResultData:
        """ Executes extension action on a server.

        Args:
            extension: extension name.
            action: action name within an extension.
            params: dictionary with parameters passed to an extension.

        Raises:
            RemoteOperationError: Communication error.

        Returns:
            Extension execution result, `returnCode==0` corresponds to success.
        """
        try:
            params_list = [[k, str(v)] for k, v in params.items()]
            extension_result = self._gql_client.execute(
                execute_extension_action_mutation,
                variable_values={
                    "extension": extension,
                    "action": action,
                    "params": params_list,
                }
            )
        except gql_exceptions.TransportServerError as error:
            if error.code:
                process_response_common(codes(error.code))
            raise
        except gql_exceptions.TransportQueryError as error:
            raise RemoteOperationError(
                error.errors if error.errors else "Unknown error occurred in the remote operation."
            ) from error

        result = ExtensionExecutionResultData.from_dict(extension_result["executeExtension"])  # pylint: disable=unsubscriptable-object
        logging.info(
            "Executed a %s / %s extension action with result code %d",
            extension, action, result.returnCode)
        return result



""" The module contains classes to represent extension-related responses from the server. """

from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel


class ExtensionParameterData(BaseModel):
    """Parameter definition for an extension action."""

    name: str
    displayName: Optional[str]
    description: Optional[str]
    dataType: str
    defaultValue: Optional[str]
    options: Optional[List[str]]


class ExtensionActionData(BaseModel):
    """Executable extension action."""

    name: str
    description: str
    experimentVariableName: str
    parameters: List[ExtensionParameterData]


class ExtensionData(BaseModel):
    """Dataclass for defining an extension"""

    name: str
    description: Optional[str]
    authors: str
    actions: List[ExtensionActionData]

    @classmethod
    def from_dict(cls, data: dict) -> ExtensionData:
        """Composes an object from a server response.

        Args:
            data: server response.

        Returns:
            Object populated with server response data.
        """
        return ExtensionData(**data)


class ExtensionExecutionResultData(BaseModel):
    """Results of extension execution"""

    returnCode: int
    stdout: str
    stderr: str
    logExperiment: str
    logFile: str

    @classmethod
    def from_dict(cls, data: dict) -> ExtensionExecutionResultData:
        """Composes an object from a server response.

        Args:
            data: server response.

        Returns:
            Object populated with server response data.
        """
        return ExtensionExecutionResultData(**data)



"""Aqueduct GraphQL Query schemas"""

from gql import gql

get_experiments_query = gql(
    """
    query GetExperiments (
        $limit: Int!,
        $offset: Int!,
        $title: String = null,
        $startDate: DateTime = null,
        $endDate: DateTime = null,
        $tags: [String!] = null
    ) {
        experiments (
            filters: {
                title: $title,
                startDate: $startDate,
                endDate: $endDate,
                tags: $tags
            }
            limit: $limit,
            offset: $offset,
        ) {
            experimentsData {
                uuid
                title
                description
                eid
                createdAt
                updatedAt
                tags
                files {
                    name
                    path
                    modifiedAt
                }
            }
            totalExperimentsCount
        }
    }
    """
)


get_experiment_query = gql(
    """
    query GetExperimentByIdentifier (
        $type: IDType!,
        $value: String!,
    ) {
      experiment (
        experimentIdentifier: {
            value: $value,
            type: $type,
        }
       ) {
        uuid
        title
        description
        eid
        createdAt
        updatedAt
        tags
        files {
            name
            path
            modifiedAt
        }
      }
    }
    """
)

get_all_tags_query = gql(
    """
    query GetAllTags (
        $limit: Int,
        $offset: Int,
        $dangling: Boolean
    ) {
    tags (
        limit: $limit,
        offset: $offset,
        filters: {
            includeDangling: $dangling
        }
    ) {
        tagsData
        totalTagsCount
    }
    }
    """
)

get_all_extensions_query = gql(
    """
    query MyQuery {
        extensions {
            name, authors, description,
            actions {
                name, description, experimentVariableName,
                parameters {
                    name
                    displayName
                    description
                    dataType
                    defaultValue
                    options
                }
            }
        }
    }
    """
)



"""Aqueduct GraphQL Mutation schemas"""

from gql import gql

create_experiment_mutation = gql(
    """
    mutation CreateExperiment (
        $title: String!,
        $description: String!,
        $tags: [String!]!
    ) {
        createExperiment (
            createExperimentInput: {
                title: $title,
                description: $description,
                tags: $tags
            }
        ) {
            uuid
            title
            description
            eid
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

update_experiment_mutation = gql(
    """
    mutation UpdateExperiment(
        $uuid: UUID!,
        $title: String,
        $description: String
    ) {
        updateExperiment (
            uuid: $uuid,
            experimentUpdateInput: {
                title: $title,
                description: $description
            }
        ) {
            uuid
            title
            description
            eid
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

add_tags_to_experiment_mutation = gql(
    """
    mutation AddTagToExperiment (
        $uuid: UUID!,
        $tags: [String!]!
    ) {
        addTagsToExperiment(
            experimentTagsInput: {
                uuid: $uuid,
                tags: $tags
            }
        ) {
            uuid
            title
            description
            eid
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

remove_tag_from_experiment_mutation = gql(
    """
    mutation RemoveTagFromExperiment (
        $uuid: UUID!,
        $tag: String!
    ) {
        removeTagFromExperiment (
            experimentTagInput: {
                uuid: $uuid,
                tag: $tag
            }
        ) {
            uuid
            title
            description
            eid
            createdAt
            updatedAt
            tags
            files {
                name
                path
                modifiedAt
            }
        }
    }
    """
)

remove_experiment_mutation = gql(
    """
    mutation RemoveTagFromExperiment (
        $uuid: UUID!
        ) {
        removeExperiment(experimentRemoveInput: {uuid: $uuid})
    }
    """
)

execute_extension_action_mutation = gql(
    """
    mutation ExecuteExtension (
        $extension: String!,
        $action: String!,
        $params: [[String!]!]!,
    ) {
        executeExtension(
            params: $params,
            extension: $extension,
            action: $action,
        ) {
            logExperiment
            logFile
            returnCode
            stderr
            stdout
        }
    }
    """
)
